import argparse
import gzip
import os
import shutil
import sys
from abacat.abacat_helper import timer_wrapper

"""
A script to rename files in assembly directory structure.

Input:
One or more assembly directories.
Output:
One or more assembly directories with the directory and all files renamed.
"""


def ls_and_decompress(assembly_dir, unzip=True):
    """
    This function lists and decompresses the files in the assembly directory.
    Must be in NCBI ftp file hierarchy. e.g.
    ftp://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/770/275/GCF_000770275.1_ASM77027v1

    Input:
    Genome directory.
    Returns:
    List of files, decompressed.
    """
    # I prefer dealing with absolute paths.
    assembly_dir = os.path.abspath(assembly_dir)

    # Quick error handling
    if not os.path.isdir(assembly_dir):
        raise FileNotFoundError

    files = [os.path.join(assembly_dir, file) for file in os.listdir(assembly_dir)]

    if unzip:
        print(f"Decompressing files in {assembly_dir}.")
        for file in files:
            if file.endswith(".gz"):
                with gzip.open(file, "r") as f_in, open(file[:-3], "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
                os.remove(file)

        # Refresh the 'files' var with unzipped names
        files = [os.path.join(assembly_dir, file) for file in os.listdir(assembly_dir)]

    return files


def parse_assembly_report(assembly_dir):
    """
    Parses information from assembly_report. Uses information to rename.

    Input:
    Genome directory.

    Returns:
    Tuple containing (organism_name, assembly_name, assembly_accession)

    TODO: Improve this function. See RefSeq/GenBank use cases. Major error handling.
    """
    files = ls_and_decompress(assembly_dir, unzip=False)

    for file in files:
        if file.endswith("assembly_report.txt"):
            assembly_report = file

    if not assembly_report:
        return "Genome report not found, please check your assembly directory.\n"

    with open(assembly_report) as report:
        r = [line for line in report.readlines()]
        for line in r:
            if line.startswith("# Genome name:"):
                assembly_name = line.strip().split(":")[1].strip()

            if line.startswith("# Organism name:"):
                organism_name = "_".join(line.strip().split(":")[1].strip().split())

            if line.startswith("# GenBank assembly accession:"):
                assembly_accession = line.strip().split(":")[1].strip()
                if "GCF" in assembly_report:
                    assembly_accession = assembly_accession.replace("GCA", "GCF")

    return assembly_name, organism_name, assembly_accession


def rename_assembly(assembly_dir, rename="organism assembly", parse_organism_name=3):
    """
    Use information from the assembly report to rename our directory and files.

    Input:
    Genome directory.

    Returns:
    Renamed assembly directory and files.

    rename:
    How the file should be renamed.

    parse_organism_name INT:
    Length of organism name epithets.

    Todo: Add renaming options, like assembly + accession, organism + accession.
          Maybe write a helper function to do this.
    """
    assembly_dir = os.path.abspath(assembly_dir)

    print(f"Renaming files in {assembly_dir}\n")

    assembly_name, organism_name, assembly_accession = parse_assembly_report(
        assembly_dir
    )
    preffix = "_".join([assembly_accession, assembly_name])
    files = ls_and_decompress(assembly_dir, unzip=False)

    # This code is unnecessary for now, but it will help with the renaming options later
    if "organism" in rename:
        # This if defines the length of the organism epithets.
        # For 3 we get Genus, species and strain (default)
        if len(organism_name.split("_")) > parse_organism_name:
            organism_name = organism_name.split("_")[:parse_organism_name]
            organism_name = "_".join(organism_name)

        fname = "_".join([organism_name, assembly_name])

    # These characters will be removed from the final filename.
    # This is because they might interefere with posix path.
    bad_chars = "\\/.?()[]{}#<>º´'"
    for char in bad_chars:
        fname = fname.replace(char, "")

    # Rename the files
    for file in files:
        if preffix in os.path.basename(file):
            new_file = os.path.basename(file).replace(preffix, fname)
            os.rename(file, os.path.join(assembly_dir, new_file))

    # Rename the directory
    os.rename(assembly_dir, os.path.join(os.path.dirname(assembly_dir), fname))
    return os.path.join(os.path.dirname(assembly_dir), fname)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Renames assembly directories.")
    parser.add_argument(
        "-i",
        "--input",
        help="Genome directory or directory containing multiple assembly directories.",
    )
    args = parser.parse_args()

    if not args.input:
        parser.print_help()
        sys.exit(0)

    @timer_wrapper
    def main():
        # Check if it is a single assembly directory or multiple:
        if "annotation_hashes.txt" in os.listdir(args.input):
            ls_and_decompress(args.input)
            rename_assembly(args.input)

        else:
            directories = os.listdir(args.input)
            success, failure = 0, 0
            for directory in directories:
                try:
                    directory = os.path.join(args.input, directory)
                    ls_and_decompress(directory)
                    new_dir = rename_assembly(directory)
                    if os.path.isdir(new_dir):
                        success += 1

                except Exception:
                    print(
                        f"Something went wrong with {directory}. Please confirm if it is a proper assembly directory."
                    )
                    failure += 1
                    pass

            print(
                f"Done. You successfully renamed {success} directories and had {failure} failures."
            )

    main()



import abacat
import pytest
import pandas
from os import path

"""
Module for testing the ANIDendrogram class and dependencies.
"""

key_file = path.join(abacat.genomes_dir, "key_file.tsv")
list_of_genomes = list(abacat.CONFIG["test_genomes"].values())
list_of_genomes = [abacat.Genome(file) for file in list_of_genomes]
output_dir = "tmp_test/"
dn = abacat.ANIDendrogram(output_dir=output_dir)

# See if FastANI binary is there
def test_pyani_install():
    assert path.isfile(abacat.CONFIG["third_party"]["fastANI"]), "Please check your FastANI binary path in the Abacat config file. It should match the `which fastANI` command."


def test_ANIDendrogram_class():
    assert type(dn) is abacat.dendrogram.ANIDendrogram


def test_no_genes_files():
    with pytest.raises(Exception):
        assert dn.make_fastani_input(list_of_genomes, kind="instances")


def test_run_prodigal():
    print("Running Prodigal for genomes. This might take a while.")
    ix = 1
    for genome in list_of_genomes:
        print(f"Running for {ix}/{len(list_of_genomes)} genomes.")
        genome.run_prodigal(quiet=True)
        ix += 1


def test_list_of_genomes():
    assert(len(list_of_genomes) == 7)


def test_make_fastani_input():
     dn.make_fastani_input(list_of_genomes=list_of_genomes, kind="instances")
     assert path.isfile(dn.fastani_input)


def test_fastani_run():
    dn.run()
    assert path.isfile(dn.fastani_output)


def test_make_ani_table():
    dn.make_ani_table()
    assert path.isfile(dn.ani_table)


def test_import_key_file():
    dn.import_key_file(key_file)
    assert type(dn.key_df) is pandas.core.frame.DataFrame


def test_make_dendrogram():
    dn.make_dendrogram(color_threshold=5, filter_rename=True)
    assert path.isfile(dn.fig_output + ".png")



import os
from abacat import timer_wrapper, CONFIG
from Bio import Entrez
from Bio import SeqIO

Entrez.email = CONFIG["email"]


class GenomeFetch:
    """
    Class GenomeFetch. Hosts accession numbers that will be queried.
    """

    def __init__(self, records_file=None):
        super(GenomeFetch, self).__init__()
        self.records_file = records_file
        self.accessions = None
        self.out_dir = CONFIG["out_dir"]
        if records_file:
            if os.path.splitext(records_file)[1] in (
                ".fna",
                ".fasta",
            ):  # Checks if it is a FASTA file.
                self.load_accessions(fasta=True)
            else:
                self.load_accessions()

    def load_accessions(self, fasta=False):
        """
        Load accession numbers as queries.
        Fasta: if True, accession numbers should be in a fasta file as sequence
        descriptions. Else, should be in a file, one per line.
        """
        if fasta:
            with open(self.records_file) as f:
                records = SeqIO.parse(f, "fasta")
                desc = ["_".join(i.description.split("_")[:2]) for i in records]
                self.accessions = [
                    Query(j, out_dir=self.out_dir) for j in {i for i in desc}
                ]  # Don't save the index.
        else:
            with open(self.records_file) as f:
                records = f.readlines()
                records = [i.strip() for i in records]
                self.accessions = [Query(i, out_dir=self.out_dir) for i in records]


class Query(object):
    """
    Query to nucletide database. Should be a valid NCBI accession number.

    It is important to note that some of the accession numbers need to be converted.
    """

    def __init__(self, repr, out_dir=None):
        super(Query, self).__init__()
        self.repr = repr
        self.out_dir = os.path.join(out_dir, repr)
        self.out_gb = os.path.join(self.out_dir, self.repr + ".gbk")
        self.out_fasta = os.path.join(self.out_dir, self.repr + ".fasta")

    def __repr__(self):
        return self.repr

    def fetch_query(self, force=False, gb=True, fasta=True):
        query = self.repr
        out_dir = self.out_dir

        print(f"Searching for query {query}.")

        if not force:
            if os.path.isfile(self.out_gb):
                print(
                    f"{self.out_gb} already exists. Use force or set a different path."
                )
                gb = False
            if os.path.isfile(self.out_fasta):
                print(
                    f"{self.out_fasta} already exists. Use force or set a different path."
                )
                fasta = False

        if not os.path.isdir(self.out_dir):
            os.mkdir(self.out_dir)

        if gb:

            @timer_wrapper
            def write_gb():
                net_handle = Entrez.efetch(
                    db="nucleotide", id=query, rettype="gb", retmode="text"
                )
                r = net_handle.read()
                with open(self.out_gb, "w") as out_handle:
                    out_handle.write(r)
                    net_handle.close()
                    print(f"Saved .gbk record for query {query} at {self.out_gb}.")

            write_gb()

        if fasta:

            @timer_wrapper
            def write_fasta():
                net_handle = Entrez.efetch(
                    db="nucleotide", id=query, rettype="fasta", retmode="text"
                )
                r = net_handle.read()
                if len(r) < 10:
                    print("Failed to find .FASTA for this entry.")
                    return
                with open(self.out_fasta, "w") as out_handle:
                    out_handle.write(r)
                    net_handle.close()
                    print(f"Saved .fasta record for query {query} at {self.out_fasta}.")

            write_fasta()



name = "abacat"

from abacat.genome import Genome, from_fasta, from_json
from abacat.abacat_helper import get_records, is_fasta, is_fasta_wrapper, timer_wrapper
from abacat.prodigal import Prodigal, run
from abacat.config import CONFIG, pathways
from abacat.deprecated import (
    prokka,
    ffn_parser,
    ls_and_decompress,
    parse_assembly_report,
    rename_assembly,
    dict_from_report,
)
from abacat.data import data_dir, genomes_dir, local_db_dir
from abacat.dendrogram import ANIDendrogram



"""
This script runs batch GenomeFetch in parallel to fetch queries.
"""

import os
import argparse
from multiprocessing.pool import ThreadPool
from abacat import CONFIG, GenomeFetch, Query
from Bio import Entrez
from operator import methodcaller

Entrez.email = CONFIG["email"]


def main(file_with_accessions):
    """
    file_with_accessions: our input with all accession numbers we want to
    fetch from NCBI.
    """
    genfet = GenomeFetch(file_with_accessions)

    return genfet.accessions


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input")
    parser.add_argument("-nf", "--nofasta", dest="fasta", action="store_false")
    parser.add_argument("-ng", "--nogb", dest="gb", action="store_false")
    parser.add_argument("-f", "--force", dest="force", action="store_true")
    parser.add_argument("-t", "--threads", default=int(os.cpu_count() / 2))
    parser.set_defaults(fasta=True, gb=True, force=False)

    args = parser.parse_args()

    acc = main(args.input)

    with ThreadPool(int(args.threads)) as p:
        f = methodcaller("fetch_query", force=args.force, fasta=args.fasta, gb=args.gb)
        p.map(f, acc)



"""
Put all assembly report information into a single table.
Usage:

python assembly_report_extractor.py <PATH TO REPORTS> <PATH OF OUTPUT DATAFRAME>

Where <PATH TO REPORTS> is a directory containing assembly reports, and
And <PATH OF OUTPUT DATAFRAME> is your output file name (.csv extension).

# TODO: Fix append method to df.

'Premature optimization is the root of all evil.' Donald Knuth.

"""

import os
import sys
import argparse
import pandas as pd
from abacat.abacat_helper import timer_wrapper


def dict_from_report(report):
    """
    Makes a dictionary from the report, which will then be converted into a dataframe.
    """
    with open(report) as f:
        r = f.readlines()
        r = r[:15]
        r = [line.strip() for line in r]
        r = [line.replace("# ", "") for line in r]
        r = [line.split(":") for line in r]
        r = [(line[0], line[1].lstrip()) for line in r if len(line) > 1]
        dict_ = dict(line for line in r)
        dict_["fname"] = os.path.basename(report).split("_assembly_report.txt")[0]

        return dict_


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="A script to put NCBI Genome reports into a csv file."
    )
    parser.add_argument(
        "-i", "--input", help="Genome report or path with assembly_reports."
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Path to output .csv file.",
        default="assembly_reports.csv",
    )
    args = parser.parse_args()

    if not args.input:
        parser.print_help()
        sys.exit(0)

    @timer_wrapper
    def main():
        if os.path.isfile(args.input):
            output = os.path.splitext(args.input)[0] + ".csv"
            dict_ = dict_from_report(args.input)
            df = pd.DataFrame(dict_, index=[0])
            if df.shape[1] < 2:
                raise Exception("Invalid input. Please check your input file.")

            df.to_csv(args.output, sep="\t", index=False)
            print(f"Converted {args.input} into {output}")

        elif os.path.isdir(args.input):
            ix = 0
            reports = [
                i for i in os.listdir(args.input) if i.endswith("assembly_report.txt")
            ]
            print(f"You have {len(reports)} reports.")
            for report in reports:
                try:
                    report = os.path.join(os.path.abspath(args.input), report)
                    dict_ = dict_from_report(report)
                    df = pd.DataFrame(dict_, index=[0])
                    if ix == 0:
                        df.to_csv(args.output, sep="\t", index=False)
                    else:
                        with open(args.output, "a") as f:
                            df.to_csv(f, sep="\t", header=False, index=False)
                    ix += 1
                except:
                    pass

            print(f"Done. Reports written to {args.output}.")

    main()



"""
Create ANI dendrograms using FastANI and SciPy.

Input is 2 or more gene or genome files.

Example usage as script:

python abacat/dendrogram.py -i my_genomes/

"""

import argparse
import subprocess
import pandas as pd
from os import path, listdir, mkdir
from matplotlib import pyplot as plt
from abacat import prodigal, CONFIG, timer_wrapper
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import dendrogram, linkage


class ANIDendrogram:
    def __init__(
        self,
        fastani_input="fastani_input.txt",
        threads=CONFIG["threads"],
        cmd=None,
        fraglen=300,
        fastani_output=None,
        output_dir="ani_output/",
        ani_table=None,
        fig_output=None,
        key_df=None,
    ):
        """
        A class to make ANI dendrograms using FastANI and SciPy.

        :param fastani_input: File with path of gene files, one per line, to give to fastANI
        :param threads: Number of threads to use with FastANI
        :param cmd: Command string
        :param fraglen: Length of fragments for FastANI. Default 300
        :param fastani_output: FastANI file name
        :param output_dir: Output directory name
        :param ani_table: Table with FastANI results
        :param fig_output: Dendrogram figure file name
        :param color_threshold: Parameter for color threshold in Dendrogram. Default is 5 for species level.
        :param key_df: Dataframe with two columns: old name and new name, to filter or rename samples in dendrogram.
        """
        super(ANIDendrogram, self).__init__()
        self.fastani_input = fastani_input
        self.threads = threads
        self.cmd = cmd
        self.fraglen = fraglen
        self.fastani_output = fastani_output
        self.output_dir = output_dir
        self.ani_table = ani_table
        self.fig_output = fig_output
        self.fastani_bin = CONFIG["third_party"]["fastANI"]
        self.df = None  # Pandas dataframe which will be saved to ani_table.
        self.key_df = key_df

    def import_key_file(self, key_file):
        self.key_df = import_and_validate_key_file(key_file)

    def make_fastani_input(self, list_of_genomes=None, kind="files"):
        """
        :param list_of_genomes: List of genomes to build file.
        :param kind: 'instances' if instances of Genome class. 'files' if list of files.
        :param run: runs Prodigal for genome if genes file present.
        :return: FastANI input with gene files, one per line.
        """

        if not path.isdir(self.output_dir):
            mkdir(self.output_dir)

        self.fastani_input = path.basename(self.fastani_input)  # Prevents from doubling the directory if running again.
        self.fastani_input = path.join(self.output_dir, self.fastani_input)

        if kind == "instances":
            gene_files = []
            for genome in list_of_genomes:
                if str(type(genome)) != "<class 'abacat.genome.Genome'>":
                    print(f"{genome} is not a valid Genome instance. Ignoring it.")
                    # raise Exception("Please pass a valid Genome instance")
                    pass
                elif not genome.files["prodigal"]["genes"]:
                    raise Exception(f"{genome} does not have a genes file. Ignoring it.")
                else:
                    gene_file = genome.files["prodigal"]["genes"]
                    gene_files.append(gene_file)
            with open(self.fastani_input, "w") as f:
                f.write("\n".join(gene_files))
        elif kind == "files":
            with open(self.fastani_input, "w") as f:
                for genome in list_of_genomes:
                    f"Writing genome to {self.fastani_input}."
                    f.write(genome + "\n")

        if path.isfile(self.fastani_input):
            with open(self.fastani_input) as f:
                for i, l in enumerate(f):
                    pass

            return f"Wrote {i + 1}/{len(list_of_genomes)} gene files to {self.fastani_input}."
        else:
            return f"Could write to {self.fastani_input}."

    def run(self):
        if not self.fastani_output:
            self.fastani_output = path.join(
                self.output_dir, f"fastani_out_{self.fraglen}"
            )

        if not self.cmd:
            self.cmd = f"{self.fastani_bin} --ql {self.fastani_input} --rl {self.fastani_input} -t {self.threads} -o {self.fastani_output} --minFraction 0"

        if self.fraglen:
            self.cmd += f" --fragLen {self.fraglen}"

        @timer_wrapper
        def run_():
            print("Running FastANI.")
            subprocess.call(self.cmd, shell=True)

        run_()
        if path.isfile(self.fastani_output):
            print(f"FastANI ran successfully!")

    def make_ani_table(self):
        """
        :return: ANI distance table from FastANI output.
        """

        # Setting the columns and reading the dataframe
        columns = [
            "Genome_A",
            "Genome_B",
            "ANI",
            "orthologous_fraction",
            "total_fragments",
        ]
        df = pd.read_csv(self.fastani_output, sep="\t", header=None)
        df.columns = columns

        # Creating the concatenated data frame
        df_ = df.copy()
        df_.rename(columns={"Genome_A": "Genome_B", "Genome_B": "Genome_A"}, inplace=True)
        df_t = pd.concat([df, df_], sort=True)
        df_t.ANI = round(df_t.ANI, 2)

        # Switching from 'long' form to a distance matrix
        df = pd.pivot_table(
            df_t, values="ANI", index=("Genome_B"), columns=("Genome_A")
        )
        del df_
        del df_t  # Don't need them anymore

        # Make sure the horizontal diagonal is symmetrical
        for i, j in zip(range(len(df)), range(len(df))):
            df.iloc[i, j] = round(99.99, 2)

        df.dropna(inplace=True)  # Drop na columns to prevent

        """
        The df is using the full path and genes file.
        This next bit replaces it with a shorter, prettier name.
        And allows to filter and rename with the key file.
        """

        df.index = [path.basename(i).replace("_prodigal_genes", "") for i in df.index]
        df.columns = [
            path.basename(i).replace("_prodigal_genes", "") for i in df.columns
        ]

        if not self.ani_table:
            self.ani_table = path.join(self.output_dir, "ani_table.txt")

        self.df = df
        print(
            f"Writing ANI distance matrix to {path.abspath(self.ani_table)}. It is {df.shape[0]} by {df.shape[1]}."
        )
        df.to_csv(self.ani_table, header=False, index=False)

    def make_dendrogram(self, color_threshold=5, filter_rename=False):
        """
        :param filter_rename: Will filter and rename using self.key_file. If there is not a key_file, will attempt to import from passed key_file.
        :param color_threshold: Color threshold to paint dendrogram branches.
        :return: Build dendrogram.
        """
        print(f"Plotting dendrogram with color threshold of {color_threshold}.")
        if not any(self.df):
            raise Exception(
                "You don't have an ANI table to make a dendrogram. Run make_ani_table() first."
            )

        table = self.df
        if filter_rename:
            if not path.isfile(filter_rename):
                self.import_key_file(filter_rename)

            table.rename(
                columns=dict(zip(self.key_df["old_name"], self.key_df["new_name"])),
                index=dict(zip(self.key_df["old_name"], self.key_df["new_name"])),
                inplace=True,
            )
            table = table.loc[
                list(self.key_df["new_name"]), list(self.key_df["new_name"])
            ]
        table = abs(table - 99.99)
        X = squareform(table)
        Z = linkage(X, method="complete", metric="cityblock", optimal_ordering=True)

        fig, ax = plt.subplots()

        augmented_dendrogram(
            Z,
            labels=table.columns,
            leaf_rotation=-90,
            color_threshold=color_threshold,
            leaf_font_size=12,
            ax=ax,
        )

        if not self.fig_output:
            self.fig_output = path.join(
                self.output_dir, f"ANI_dendrogram_ct_{color_threshold}"
            )
        plt.savefig(self.fig_output, bbox_inches="tight")
        if path.isfile(self.fig_output + ".png"):
            print(f"Generated dendrogram at {path.abspath(self.fig_output)}.")


def augmented_dendrogram(*args, **kwargs):

    ddata = dendrogram(*args, **kwargs)

    if not kwargs.get("no_plot", False):
        for i, d in zip(ddata["icoord"], ddata["dcoord"]):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > 1.5:
                plt.plot(x, y, "ro")
                plt.annotate(
                    "%.3g" % y,
                    (x, y),
                    xytext=(0, -8),
                    textcoords="offset points",
                    va="top",
                    ha="center",
                )

    return ddata


def import_and_validate_key_file(key_file):
    """
    :param key_file: Validates key file to make sure it will work.
    :return: Pandas dataframe with key file
    """
    with open(key_file) as f_:
        r = [i.split() for i in f_]

    for i in r:
        if len(i) != 2:
            raise Exception(
                "Invalid key file. Must be a tab delimited file with two columns, old name and new name."
            )

    r = pd.DataFrame(r)
    r.columns = ("old_name", "new_name")

    return r


if __name__ == "__main__":

    @timer_wrapper
    def main():
        parser = argparse.ArgumentParser(
            description="Plot a dendrogram based on ANI distances."
        )
        parser.add_argument(
            "-i",
            "--input",
            help="Input genomes. "
            "May be a file containing the path to each genome, one per line, or a folder containing genomes. "
            "Make sure to flag genes=True if using file containing gene sequences "
            "(otherwise will attempt to run Prodigal before).",
        )
        parser.add_argument(
            "-o",
            "--output",
            help="Path to the output directory. Will create 4 files:\n"
            "1. FastANI input file (skipped if input is file containing one genome per line).\n"
            "2. FastANI output file (tab-delimited).\n"
            "3. ANI distance matrix.\n"
            "4. Dendrogram plot.\n",
            default="./fastani_output",
        )
        parser.add_argument(
            "-g",
            "--genes",
            help="Flag to assign whether we are providing whole genome (False) or only genes (True). "
            "If False, will run Prodigal to predict genes before running FastANI (default: False).",
            type=bool,
            default=False,
        )
        parser.add_argument(
            "-k",
            "--keys",
            help="Use this parameter to rename or filter your samples. "
            "Must be a space or tab delimited file of <OLD-NAME> <NEW NAME>."
            "Samples not in this file will be filtered out.",
            default=None,
        )
        parser.add_argument(
            "-ct",
            "--color_threshold",
            help="Color threshold to paint branches of dendrogram. Default is 5 (species level).",
            type=int,
            default=5,
        )
        parser.add_argument(
            "-fl",
            "--fragLen",
            help="Fragment length to give to FastANI. Default is 300.",
            type=int,
            default=300,
        )
        parser.add_argument(
            "-t",
            "--threads",
            help="Number of threads to use. Default is set in Abacat config file.",
            type=int,
            default=CONFIG["threads"],
        )
        parser.add_argument(
            "-s",
            "--skip",
            help="Skip FastANI run. Only plot figure (must provide FastANI output).",
            type=bool,
            default=False,
        )

        args = parser.parse_args()

        # Parsing args.input
        try:
            if path.isdir(args.input):
                input_files = [
                    path.abspath(path.join(args.input, i)) for i in listdir(args.input)
                ]
                input_files = [
                    i for i in input_files if i.endswith((".fna", ".fasta", ".fa"))
                ]  # Only catch some extensions.
                if args.genes:
                    input_files = [i for i in input_files if "prodigal_genes" in i]
                else:
                    input_files = [
                        i for i in input_files if "prodigal" not in i
                    ]  # Skip previous Prodigal files
            elif path.isfile(args.input):
                if not args.genes:
                    with open(args.input) as f:
                        input_files = [path.abspath(i.strip()) for i in f.readlines()]
                else:
                    input_files = (
                        args.input
                    )  # If it is already a genes file, it is proper for FastANI. No need to write another.
        except FileNotFoundError as error:
            raise (error)
            input_files = None  # This line is only to suppress warnings.

        # Parsing args.ouput
        print(f"Set {path.abspath(args.output)} as output directory.")
        if not path.isdir(args.output):
            mkdir(args.output)
        fastani_input = path.join(args.output, "fastani_input.txt")

        # Parsing args.genes -- Prodigal runs
        if not args.genes:
            print(f"You have {len(input_files)} genomes to be processed:\n")
            print("\n".join((path.basename(i) for i in input_files)), "\n")
            print(f"Starting gene prediction.\n")
            ix, gene_files = 1, []
            for file in input_files:
                print(f"Processing file {ix}/{len(input_files)}.")
                genes = prodigal.run(file, output=path.dirname(file), quiet=True, print_files=False)
                gene_files.append(genes["genes"])
                ix += 1
            with open(fastani_input, "w") as f:
                f.write("\n".join(gene_files))
        else:
            with open(fastani_input, "w") as f:
                f.write("\n".join(input_files))

        ani = ANIDendrogram(
            output_dir=args.output,
            fastani_input=fastani_input,
            fastani_output=path.join(
                args.output, f"fastani_out_{args.fragLen}.tsv"
            ),
            threads=args.threads,
        )
        if not args.skip:
            ani.run()
        ani.make_ani_table()
        if args.keys:
            print(f"Key file set as {path.abspath(args.keys)}.")
            ani.make_dendrogram(
                color_threshold=args.color_threshold, filter_rename=args.keys
            )
        else:
            ani.make_dendrogram(color_threshold=args.color_threshold)

    # https://www.youtube.com/watch?v=odeHP8N4LKc
    main()



from Bio import SeqIO
import datetime
import time
import logging


def is_fasta(file):
    """
    Check whether a given file is a valid FASTA file.
    https://stackoverflow.com/questions/44293407/how-can-i-check-whether-a-given-file-is-fasta
    """
    with open(file) as handle:
        fasta = SeqIO.parse(handle, "fasta")
        return any(fasta)


def is_fasta_wrapper(func):
    """
    A decorator wrapper for functions which need checking for a valid FASTA.
    """

    def wrapper(*args, **kwargs):
        try:
            if not is_fasta(args[0]):
                raise Exception(
                    "Your file is not valid. Please check if it is a valid FASTA file."
                )
            return func(*args, **kwargs)
        except IndexError:
            logging.error("Please specify an input file.", exc_info=True)

    return wrapper


def timer_wrapper(func):
    """
    A wrapper to time the execution time of our functions.

    Example usage:

    if __name__ == "__main__":
        *define argparse here*

        @timer_wrapper
        def main():
            '''
            This is your main function
            '''

        main()
    """

    def wrapper(*args, **kwargs):
        start = time.time()

        func(*args, **kwargs)

        end = time.time()
        delta = str(datetime.timedelta(seconds=end - start))
        logging.info(f"Took {delta}")

    return wrapper


@is_fasta_wrapper
def get_records(fasta_file, kind="gen"):
    with open(fasta_file) as f:
        if kind == "gen":
            records = SeqIO.parse(fasta_file, format="fasta")
        elif kind == "list":
            records = list(SeqIO.parse(fasta_file, format="fasta"))
        elif kind == "dict":
            records = SeqIO.to_dict(SeqIO.parse(fasta_file, format="fasta"))
        else:
            logging.info(f"Specified {kind} kind. Please specify a valid kind.")

    return records



"""
A script to parse Prokka outputs.

Input:
Directory output of Prokka.
"""
import argparse
import os
import sys
from Bio import SeqIO
from abacat.abacat_helper import is_fasta_wrapper, timer_wrapper


@is_fasta_wrapper
def ffn_parser(fasta_file, write=True):
    """
    Scans Prokka output .ffn file and creates files with SSU seqs,
    known proteins and hypothetical proteins.

    # TODO: add RNA support (tRNA, ribonucleases, etc.)
    """

    records = SeqIO.parse(fasta_file, "fasta")
    ssu, known_prots, hypothetical_prots = (
        ("SSU", []),
        ("known_prots", []),
        ("hypothetical_prots", []),
    )

    # Generate our output file name.
    output = os.path.splitext(os.path.abspath(fasta_file))[0]

    for seq in records:
        if "16S ribosomal RNA" in seq.description:
            ssu[1].append(seq)
        elif "hypothetical" in seq.description:
            hypothetical_prots[1].append(seq)
        else:
            known_prots[1].append(seq)

    for seq_type, sequences in (ssu, known_prots, hypothetical_prots):
        if len(sequences) >= 1:
            print(f"Found {len(sequences)} {seq_type} sequences in {fasta_file}.")
            if write:
                output_ = output + f"_{seq_type}.fna"
                with open(output_, "w") as f:
                    SeqIO.write(sequences, f, "fasta")
                    print(f"Wrote {len(sequences)} {seq_type} sequences to {output_}.")

    return ssu[1], known_prots[1], hypothetical_prots[1]


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="A script to parse Prokka output files."
    )
    parser.add_argument(
        "-i", "--input", help="Directory contaning Prokka output files."
    )

    args = parser.parse_args()

    if not args.input:
        parser.print_help()
        sys.exit(0)

    @timer_wrapper
    def main():
        if not os.path.isdir(args.input):
            raise Exception("Your directory wasn't found.")

        for file in os.listdir(args.input):
            file = os.path.join(os.path.abspath(args.input), file)
            if file.endswith(".ffn"):
                ffn_parser(file)

    main()



import os
import json
import subprocess
from pathlib import Path
from abacat.data import data_dir, genomes_dir, local_db_dir

abacat_path = os.path.dirname(__file__)
external_db_path = os.path.join(Path.home(), "Bio/db")


def db(dir_name, file_name, external=False):
    """
    :param dir_name: Directory name inside the $db_path variable
    :param file_name: File name inside $dir_name. Usually a fasta file.
    :param external: True if the database is not packaged within Abacat
    :return: Joined path of db_path, dir_name, file_name
    """
    if not external:
        db = os.path.join(local_db_dir, dir_name, file_name)
    else:
        db = os.path.join(external_db_path, dir_name, file_name)
    # if not os.path.isfile(db):
    #    raise Exception(f"Database not found at {db}. Please check if it is a valid path.")
    return db


def load_phenotyping(phenotyping_json):
    with open(phenotyping_json) as f:
        phenotyping = json.load(f)

    return phenotyping


def get_third_party_bins(binaries=("fastANI", "seqstats")):
    """
    Find path of third party binaries
    :param binaries: List or tuple with name of binaries.
    :return: dict with keys as the name of the binaries, values as the path in the system.
    """
    bins = dict()
    def get_bin(bin_name):
        out = subprocess.Popen(["which", bin_name], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, stderr = out.communicate()
        return Path(stdout.strip().decode("utf-8"))

    for bin in binaries:
        bins[bin] = get_bin(bin)

    return bins


CONFIG = {
    "db": {
        "COG": db("COG", "prot2003-2014.fa", external=True),
        "megares": db("megares", "megares_database_v1.01.fasta"),
        "phenotyping": db("phenotyping", "phenotyping.fasta"),
        "pathways": db("phenotyping", "pathways.json"),
    },
    "third_party": get_third_party_bins(),  # Docker config
    "threads": int(os.cpu_count() / 2),
    "blast": {"evalue": 10 ** -20},
    "test_genomes": {
        "Staphylococcus aureus CA15": os.path.join(
            genomes_dir, "GCF_001021895.1_ASM102189v1_genomic.fna"
        ),
        "Staphylococcus aureus KUN1163": os.path.join(
            genomes_dir, "GCF_008619075.1_ASM861907v1_genomic.fna"
        ),
        "Staphylococcus epidermidis ATCC 12228": os.path.join(
            genomes_dir, "GCF_000007645.1_ASM764v1_genomic.fna"
        ),
        "Synechococcus elongatus PCC 6301": os.path.join(
            genomes_dir, "GCF_000010065.1_ASM1006v1_genomic.fna"
        ),
        "Synechococcus elongatus PCC 7942": os.path.join(
            genomes_dir, "GCF_000012525.1_ASM1252v1_genomic.fna"
        ),
        "Prochloroccocus marinus MIT 9313": os.path.join(
            genomes_dir, "GCF_000011485.1_ASM1148v1_genomic.fna"
        ),
        "Escherichia coli ATCC 1175": os.path.join(
            genomes_dir, "GCF_003697165.2_ASM369716v2_genomic.fna"
        ),
    },
    "data_dir": data_dir,
}

pathways = load_phenotyping(CONFIG["db"]["pathways"])



"""
A file containing our main classes and functions.
# TODO: Create class for sets (geneset, protset)
"""

import os
import json
import logging
import subprocess
import pandas as pd
from Bio import SeqIO
from Bio.Blast import NCBIXML
from Bio.Blast.Applications import (
    NcbiblastnCommandline,
    NcbiblastpCommandline,
    NcbiblastxCommandline,
)
from abacat.abacat_helper import get_records, is_fasta, is_fasta_wrapper, timer_wrapper
from abacat.prodigal import Prodigal
from abacat.deprecated import prokka
from abacat.config import CONFIG, pathways

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)

logger = logging.getLogger(__name__)


class Genome:
    """
    Genome, a class containing bacterial genome data and methods.

    # TODO: seqstats or Prodigal basic info (length, GC content) for metadata
    """

    def __init__(self, contigs=None, prodigal=False, name=None, directory=None):
        super(Genome, self).__init__()
        self.name = name
        self.directory = directory
        self.seqstats = None
        self.files = dict()
        self.geneset = dict()
        self.protset = dict()
        self.pathways = None

        if contigs:
            self.load_contigs(contigs)

        if prodigal:
            self.load_prodigal()

    """
    Sequence stats methods.
    """

    def print_seqstats(self):
        if self.seqstats:
            print("Sequence stats:\n")
            for key, value in self.seqstats.items():
                print(f"{key}\t\t\t{value}")
        else:
            try:
                self.load_seqstats()
                self.seqstats
            except:
                logger.error(
                    "Tried loading seqstats, but an error occurred.", exc_info=True
                )
                raise

    def fnoseqs(self, kind="prodigal", seqs="genes"):
        """
        Fast check of number of sequences in file.
        """
        with open(self.files[kind][seqs]) as f:
            records = SeqIO.parse(f, "fasta")
            len_ = sum(1 for x in records)

        return len_

    def sseqs(self, kind="prodigal", seqs="genes"):
        """
        Fast check of size of sequences file.
        from https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
        """
        size = os.stat(self.files[kind][seqs]).st_size

        def format_bytes(size):
            power = 2 ** 10
            n = 0
            power_labels = {0: "", 1: "k", 2: "m", 3: "g", 4: "t"}
            while size > power:
                size /= power
                n += 1

            return f"{round(size, 2)} {power_labels[n]}B"

        return format_bytes(size)

    def valid_contigs(self, quiet=True):
        try:
            any(self.files["contigs"])
        except Exception as e:
            logger.error("Must specify input contigs file.")
        else:
            if not quiet:
                logger.info("Your contig file is a valid FASTA file.")
            return True

    """
    Methods to load data into class instances.
    """

    def load_contigs(self, contigs):
        """
        This loads the contigs input file. I want to make it a constructor method.
        """
        contigs = os.path.abspath(contigs)

        # Must use class decorator, same as in Prodigal class.
        if not is_fasta(os.path.abspath(contigs)):
            raise Exception(
                "Your file is not valid. Please check if it is a valid FASTA file."
            )
        else:
            self.files["contigs"] = os.path.abspath(contigs)
            logger.info(f"Contigs file set as {contigs}")
            self.directory = os.path.dirname(self.files["contigs"])
            logger.info(f"Directory set as {self.directory}")
            self.name = os.path.splitext(os.path.basename(contigs))[0]
            logger.info(f"Name set as {self.name}")

    def load_seqstats(self):
        if not self.files["contigs"]:
            raise Exception(
                "Your Genome doesn't have an input file! Please provide one."
            )

        self.seqstats = dict()

        try:
            stats = subprocess.check_output(
                [CONFIG["third_party"]["seqstats"], self.files["contigs"]]
            ).decode("utf-8")
            stats = stats.split("\n")
            for n in stats:
                n = n.split(":")
                if len(n) == 2:
                    try:
                        self.seqstats[n[0]] = float(n[1].strip().replace(" bp", ""))
                    except (ValueError, IndexError) as error:
                        logger.error(
                            "Something is wrong with the seqstats output. Please check your seqstats command.",
                            exc_info=True,
                        )
            self.seqstats
        except FileNotFoundError:
            raise

    def load_prodigal(
        self, prodigal_out=None, load_geneset=True, load_protset=True, print_=False
    ):
        """
        Attach Prodigal results to class object.

        Input:
        prodigal_out is the output folder generated by the Prodigal function.
        """
        if prodigal_out:
            input = os.path.abspath(prodigal_out)
        else:
            input = self.directory
        # This dictionary is the same as output_files from the prodigal.py script.
        self.files["prodigal"] = dict()

        try:
            os.path.isdir(input)
        except FileNotFoundError:
            print(f"{prodigal_out} directory not found.")

        for file_ in os.listdir(input):
            if file_.startswith(self.name):
                file_ = os.path.join(input, file_)
                if file_.endswith("_genes.fna"):
                    self.files["prodigal"]["genes"] = file_
                elif file_.endswith("_proteins.faa"):
                    self.files["prodigal"]["proteins"] = file_
                elif file_.endswith("_cds.gbk"):
                    self.files["prodigal"]["cds"] = file_
                elif file_.endswith("_scores.txt"):
                    self.files["prodigal"]["scores"] = file_
                else:
                    if print_:
                        logger.debug(
                            f"{file_} apparently is not a Prodigal output file. Ignoring it."
                        )
                    pass
        logger.info(f"Loaded Prodigal files for {self.name}.")

        if load_geneset:
            self.load_geneset()
        if load_protset:
            self.load_protset()

    def load_geneset(self, kind="prodigal", records="dict"):
        """
        Loads gene sets unto Genome.geneset.
        Uses the 'genes' key from the files[kind] dictionary.
        """
        if kind == "prodigal":
            # Origin is the file from which the set came from
            origin = self.files["prodigal"]["genes"]
            try:
                self.geneset["prodigal"] = dict()
                self.geneset["prodigal"]["records"] = get_records(origin, kind=records)
                self.geneset["prodigal"]["origin"] = origin

            except Exception:
                raise
        elif kind in CONFIG["db"].keys():
            origin = self.files[kind]["annotation"]
            try:
                self.geneset[kind] = dict()
                self.geneset[kind]["records"] = get_records(origin, kind=records)
                self.geneset[kind]["origin"] = origin
            except Exception:
                raise
        else:
            logger.error(
                f"Passed {kind} kind of geneset input. Please specify a valid input from either an annotation or Prodigal file.",
                exc_info=True,
            )

        # Maybe change this if/else block later.
        if self.geneset[kind]:
            if records != "gen":
                logger.info(
                    f"Loaded gene set from {kind.capitalize()} data. It has {len(self.geneset[kind]['records'])} genes."
                )
            else:
                logger.info(f"Loaded gene set from {kind.capitalize()} data.")
        else:
            print(f"No gene set found in {origin}.")

    def load_protset(self, kind="prodigal", records="dict"):
        """
        Loads protein sets unto Genome.protset.
        Uses the 'protein' key from the files[kind] dictionary.
        """
        if kind == "prodigal":
            origin = self.files["prodigal"]["proteins"]
            try:
                self.protset["prodigal"] = dict()
                self.protset["prodigal"]["records"] = get_records(origin, kind=records)
                # TODO: attach origin to a variable (stated by 'kind')
                self.protset["prodigal"]["origin"] = origin

            except Exception:
                raise
        elif kind == "prokka":
            origin = self.files["prokka"]["proteins"]
            try:
                self.protset["prokka"] = dict()
                self.protset["prokka"]["records"] = get_records(origin, kind=records)
                self.protset["prokka"]["origin"] = origin
            except Exception:
                raise

        if self.protset[kind]:
            if records == "list":
                logger.info(
                    f"Loaded protein set from {kind.capitalize()} data. It has {len(self.protset[kind]['records'])} proteins."
                )
            else:
                logger.info(f"Loaded protein set from {kind.capitalize()} data.")
        else:
            print(f"No proteins set found in {origin}.")

    """
    Run methods. Possess the @timer_wrapper decorator to measure runtime.
    """

    @timer_wrapper
    def df_prodigal(self, kind="gene"):
        if not self.geneset["prodigal"]:
            try:
                self.load_geneset()
            except Exception("Please load your Prodigal geneset."):
                raise

        self.geneset["prodigal"]["df"] = pd.DataFrame(
            columns=(
                "id",
                "start",
                "stop",
                "strand",
                "id_",
                "partial",
                "start_type",
                "rbs_motif",
                "rbs_spacer",
                "gc_cont",
            )
        )

        def extract_row(record):
            row = record.description.split(";")
            row = row[0].split(" # ") + row[1:]
            row = [i.split("=")[-1] for i in row]
            return row

        for record in self.geneset["prodigal"]["records"]:
            # This one liner appends each record to the end of the dataframe.
            self.geneset["prodigal"]["df"].loc[
                len(self.geneset["prodigal"]["df"])
            ] = extract_row(record)

    @timer_wrapper
    def run_prodigal(self, quiet=True, load_sets=["gene", "prot"]):
        """
        Check for contigs file, run Prodigal on file.
        """
        self.valid_contigs(quiet)
        input = self.files["contigs"]
        logger.info(
            f"Starting Prodigal. Your input file is {input}. Quiet setting is {quiet}."
        )
        p = Prodigal(input, output=self.directory, quiet=quiet)
        self.files["prodigal"] = p.run()
        if "gene" in load_sets:
            self.load_geneset()
        if "prot" in load_sets:
            self.load_protset()

    @timer_wrapper
    def run_prokka(self, quiet=True, load_sets=["gene", "prot"], **kwargs):
        """
        Check for contigs file, run Prokka on file.
        """
        self.valid_contigs(quiet)
        input = self.files["contigs"]
        logger.info(
            f"Starting Prokka. Your input file is {input}. Quiet setting is {quiet}."
        )
        prokka_out = prokka(input, **kwargs)
        self.files["prokka"] = prokka_out
        if "gene" in load_sets:
            self.load_geneset("prokka")
        if "prot" in load_sets:
            self.load_protset("prokka")

    @timer_wrapper
    def blast_seqs(self, db, blast="n", evalue=CONFIG["blast"]["evalue"]):
        """
        Blasts geneset.
        :param db: From config.py db
        :param evalue: evalue to use in Blast
        """
        try:
            db_path = CONFIG["db"][db]
        except KeyError:
            logger.error(
                f"Choose a valid database from {CONFIG['db'][db]}.", exc_info=True
            )
        query = self.files["prodigal"]["genes"]
        out = os.path.join(self.directory, self.name + f"_{db}_blast.xml")
        self.files[db] = dict()
        self.files[db]["xml"] = out
        logger.info(f"Blasting {self.name} to {out}.")

        def blast_method(*args, **kwargs):
            if blast in ("n", "nucl", "nucleotide", "blastn"):
                blast_cmd = NcbiblastnCommandline(*args, **kwargs)
            elif blast in ("p", "prot", "protein", "blastp"):
                blast_cmd = NcbiblastpCommandline(*args, **kwargs)
            elif blast in ("x", "blastx"):
                blast_cmd = NcbiblastxCommandline(*args, **kwargs)
            else:
                raise Exception(
                    "Choose a valid option from 'blastn', 'blastp' or 'blastx'."
                )

            return blast_cmd

        blast_cmd = blast_method(
            query=query,
            db=db_path,
            evalue=evalue,
            out=out,
            outfmt=5,
            num_alignments=5,
            num_threads=CONFIG["threads"],
        )
        stdout, stderr = blast_cmd()
        self.parse_xml_blast(db)

    def parse_xml_blast(self, db, write_hits=True):
        """
        Blasts XML out
        :return:
        """
        hits = []
        with open(self.files[db]["xml"]) as f:
            blast_records = NCBIXML.parse(f)
            for i in blast_records:
                if i.alignments:
                    i.id = i.query.split(" #")[0]
                    i.query = " ".join((i.id, i.alignments[0].hit_def))
                    hits.append(i)
        logger.info(f"Found {len(hits)} hits.\n")

        self.geneset[db] = dict()
        self.geneset[db][
            "origin"
        ] = f"Blast of {self.files['prodigal']['genes']} to {CONFIG['db'][db]}."
        self.geneset[db]["records"] = list()

        for i in hits:
            annotation = self.geneset["prodigal"]["records"][i.id]
            annotation.description = i.query
            self.geneset[db]["records"].append(annotation)
            print(i.query)

        if write_hits:
            out_f = os.path.join(self.directory, self.name + f"_{db}.fasta")
            out_h = os.path.join(self.directory, self.name + f"_{db}.hits")
            self.files[db]["annotation"] = out_f
            self.files[db]["hits"] = out_h
            with open(out_f, "w") as f:
                SeqIO.write(self.geneset[db]["records"], f, "fasta")

            with open(out_h, "w") as f:
                for i in [j.description for j in self.geneset[db]["records"]]:
                    f.write(i + "\n")

            logger.info(
                f"Wrote {len(self.geneset[db]['records'])} annotated sequences to {out_f}."
            )

    def to_json(self, out_path=None):
        """
        Writes a json output of our genome object, that can be load back into Abacat.
        We filter the geneset and protset keys because they would take too much space,
        loading them back when we import the json obj.
        """
        json_out = dict()
        for key, value in self.__dict__.items():
            if key not in ("geneset", "protset"):
                json_out[key] = value

        if not out_path:
            out_path = os.path.join(self.directory, self.name + ".json")
        with open(out_path, "w") as f:
            json.dump(json_out, f, indent=3)
        logger.info(f"Wrote json file of {self.name} to {out_path}.")

    def run_pathways(self, info=True, evalue=10 ** -3):
        """
        Takes the phenotyping geneset records and checks them against the pathways object
        from the CONFIG module.
        :return: pathway genes, a dict containing pathways as keys and identified records as values.
        """
        if "phenotyping" not in self.files.keys():
            logger.info("Phenotyping files not found. Running BLAST now.")
            self.blast_seqs(db="phenotyping", blast="blastx", evalue=evalue)
        elif "phenotyping" not in self.geneset.keys():
            logger.info("Phenotyping records not found. Loading from BLAST out.")
            self.load_geneset(kind="phenotyping")

        self.pathways = dict()
        for (
            k,
            v,
        ) in (
            pathways.items()
        ):  # Note that this is the 'pathways' imported from config.py
            self.pathways[k] = []
            for gene in self.geneset["phenotyping"]["records"]:
                desc = gene.description.split()[1].split(".")[1]
                if desc in v:
                    self.pathways[k].append(gene.description)
            if info:
                logger.info(f"Found {len(self.pathways[k])} genes for {k}.")


@is_fasta_wrapper
def from_fasta(fasta_file, run_prodigal=False, load_prodigal=False):
    """
    A function to load assemblies and run Prodigal directly.

    Input:
    A valid contigs file.

    Returns:
    An Genome object.
    """
    genome = Genome()

    logger.info(f"Loading contigs file from {fasta_file}.")
    genome.load_contigs(fasta_file)
    if run_prodigal:
        print(f"Running Prodigal for {genome.name}.")
        genome.run_prodigal()

    if load_prodigal:
        genome.load_prodigal()

    return genome


def from_json(json_file):
    """
    :param json_file: A json file like the one exported from Genome.to_json()
    :return: A genome instance.
    """
    logger.info(f"Loading genome from {json_file}.")
    with open(json_file) as f:
        j = json.load(f)

    g = Genome(name=j["name"], directory=j["directory"])

    for attribute in ("files", "seqstats", "pathways"):
        if attribute in j.keys():
            setattr(g, attribute, j[attribute])
            logger.info(f"Loaded {attribute} to {g.name}.")

    g.load_geneset()
    g.load_protset()

    del j
    return g



from .genome_fetcher import GenomeFetch, Query

genfet = GenomeFetch("/Users/viniWS/storage/neorefs/rev6/test/test_rev6.fasta")

assemblies = dict()

for acc in genfet.accessions:
    assemblies[acc.repr] = GenomeFetch(acc.out_fasta)

for k, v in assemblies.items():
    v.load_prodigal()



"""
- access downloaded GBK file.
- extract assembly ID.
"""

import os
import argparse
from Bio import SeqIO


def find_gbk_files(input_dir):
    gbk_files = []
    for dirs, subdirs, files in os.walk(input_dir):
        for file in files:
            if os.path.splitext(file)[1] == ".gbk":
                gbk = os.path.join(dirs, file)
            if os.path.splitext(file)[1] == ".fasta":
                fasta = True
            else:
                fasta = False
            if fasta == False:
                gbk_files.append(gbk)
            else:
                pass

    return gbk_files


def extract_assembly_accession(list_of_gbk_files, write_to="accessions.txt"):
    acc = []
    not_found = []
    not_found_p = os.path.splitext(write_to)[0] + "_not_found.txt"
    for gbk in list_of_gbk_files:
        with open(gbk) as f:
            gbk_ = os.path.basename(os.path.splitext(gbk)[0])
            record = SeqIO.parse(f, "genbank")
            try:
                acc_id = [i for i in next(record).dbxrefs if "Genome" in i][0].split(
                    "Genome:"
                )[1]
                acc.append((gbk_, acc_id))
            except IndexError:
                not_found.append(gbk_)
                print(gbk_)
                pass

    if write_to:
        with open(write_to, "w") as f:
            for i in acc:
                f.write(":".join(i) + "\n")

        if len(not_found) >= 1:
            with open(not_found_p, "w") as f:
                for i in not_found:
                    f.write(i)

        if os.path.isfile(write_to):
            print(f"Wrote {len(acc)} accessions to {write_to}.")

        if os.path.isfile(not_found_p):
            print(f"Wrote {len(not_found)} not found {not_found_p}.")

    return acc


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Used to extract Genome accessions from a list of .gbk files inside directories."
    )
    parser.add_argument("-i", "--input")
    parser.add_argument("-o", "--output")
    args = parser.parse_args()

    gbk = find_gbk_files(args.input)
    extract_assembly_accession(gbk, write_to=args.output)



import abacat
import pytest
from os import path

"""
Module for testing the Genome class and related methods.
"""
input_contigs = abacat.CONFIG["test_genomes"]["Staphylococcus aureus CA15"]
g = abacat.Genome()

# If you change your test file, be sure to modify this.
assert_values = {
    "load_contigs": {
        "name": path.splitext(path.basename(input_contigs))[0],
        "directory": path.dirname(input_contigs),
        "files": {"contigs": path.abspath(input_contigs)},
    },
    "sseqs": "2.64 mB",
    "seqstats": {
        "Total n": 2.0,
        "Total seq": 2864278.0,
        "Avg. seq": 1432139.0,
        "Median seq": 1432139.0,
        "N 50": 2839253.0,
        "Min seq": 25025.0,
        "Max seq": 2839253.0,
    },
}


def test_genome_class():
    """
    :return: asserts if the class instantiates correctly.
    """
    assert type(g) is abacat.genome.Genome


def test_load_contigs():
    """
    :return: asserts if a valid contig loads correctly as a class attribute.
    """
    g.load_contigs(input_contigs)
    errors = []

    for key, value in assert_values["load_contigs"].items():
        if getattr(g, key) != value:
            errors.append(key)
    assert (
        not errors
    ), f"Errors in the following attributes: {errors} when invoking Genome.load_contigs()"


def test_invalid_contigs():
    """
    :return: asserts that we return an exception when loading an invalid file as contigs file.
    """
    h = abacat.Genome()
    with pytest.raises(Exception):
        assert h.load_contigs(abacat.CONFIG["db"]["pathways"])


def test_seqstats():
    """
    :return: Tests whether seqstats values return true.
    """
    g.load_contigs(input_contigs)
    g.load_seqstats()
    errors = []
    for (key1, value1), (key2, value2) in zip(
        g.seqstats.items(), assert_values["seqstats"].items()
    ):
        if key1 != key2 or value1 != value2:
            errors.append(f"{key1} value is {value1} but should be {value2}.")

    assert not errors, f"Errors in the following keys:\n{errors}"


def test_run_prodigal():
    """
    :return: Runs Prodigal for our genome.
    """
    g.load_contigs(input_contigs)
    g.run_prodigal()
    errors = []
    for key, value in g.files["prodigal"].items():
        if not path.isfile(value):
            errors.append(f"{key} file was not found at {value}.")

    assert not errors, f"Errors in the following keys:\n{errors}."


def test_sseqs():
    """
    :return: tests for Genome.sseqs, sequence size method.
    """
    assert g.sseqs() == assert_values["sseqs"]


def test_load_prodigal():
    """
    :return: If existing Prodigal data is loaded correctly
    """
    g = abacat.Genome(input_contigs)
    g.load_prodigal()
    errors = []
    for key, value in g.files["prodigal"].items():
        if not path.isfile(value):
            errors.append(f"Couldn't create Prodigal {key} file.")

    assert not errors, f"Errors in the following files:\n{errors}."


def test_blast_seqs_megares():
    """
    :return: Blasts the Genome object against the Megares database.
    """
    g.blast_seqs("megares")
    errors = []
    for key, value in g.files["megares"].items():
        if not path.isfile(value):
            errors.append(f"Could find {key} file at {value}.")

    assert not errors, f"Errors with the following files:\n{errors}"


def test_to_json():
    """
    :return: Asserts the Genome.to_json() method is working.
    """
    g.to_json()
    assert path.isfile(path.join(g.directory, g.name + ".json"))


# Methods from the genome.py module but not the Genome class
def test_from_fasta():
    """
    :return: Tests from_fasta import method.
    """
    h = abacat.from_fasta(input_contigs)
    errors = []

    for key, value in assert_values["load_contigs"].items():
        if getattr(h, key) != value:
            errors.append(key)
    assert (
        not errors
    ), f"Errors in the following attributes: {errors} when invoking Genome.load_contigs()"


def test_from_json():
    """
    :return: Tests from_json import method.
    """
    h = abacat.from_json(path.join(g.directory, g.name + ".json"))
    assert type(h) is abacat.genome.Genome




"""
A script to call Prodigal to predict both genes and proteins.

Input:
File or folder with contigs in fasta or fna format.

Output:
Separate files or folders with genes and proteins (default) or only one of each.
Includes log and stats file (errors, exceptions, how many genes and proteins)
When output is multiple, includes single log file.

Example usage:

    python prodigal.py -i contigs/
    python prodigal.py -i contigs.fasta
"""

import os
import sys
import argparse
import subprocess
from abacat.abacat_helper import is_fasta, is_fasta_wrapper, timer_wrapper


class Prodigal:
    """
    This class will hold Prodigal run data.

    Calls Prodigal on an input file.

    Input:
    Valid FASTA file.
    Output:
    Genes (.fna), proteins (.faa), gene scores (.txt), gbk file (.gbk).

    """

    def __init__(self, contigs, output=None, quiet=False, scores=False):
        super(Prodigal, self).__init__()
        self.name = None
        is_fasta(contigs)
        self.contigs = contigs  # an assembled genome contigs file.
        self.quiet = quiet
        self.finished = None
        self.scores = scores

        if not output:
            output = os.path.join(
                os.getcwd(),
                os.path.basename(os.path.splitext(self.contigs)[0]) + "_prodigal",
            )
        else:
            output = os.path.join(
                output,
                os.path.basename(os.path.splitext(self.contigs)[0]) + "_prodigal",
            )
        self.output = os.path.join(os.path.abspath(output), output.split("/")[-1])
        self.output_files = {
            "genes": output + "_genes.fna",
            "proteins": output + "_proteins.faa",
            "cds": output + "_cds.gbk",
        }

        self.cmd = f"prodigal -i {self.contigs} -a {self.output_files['proteins']} \
                    -d {self.output_files['genes']} -o {self.output_files['cds']}"

        if self.scores:
            self.output_files["scores"] = output + "_scores.txt"
            self.cmd += f" -s {self.output_files['scores']}"
        if self.quiet:
            self.cmd += " -q"

    def run(self, print_files=False):
        is_fasta_wrapper(self.contigs)
        subprocess.call(self.cmd, shell=True)

        if all(os.path.isfile(value) for _, value in self.output_files.items()):
            self.finished = True
            if print_files:
                print(f"Created files at {self.output}:")
                for _, v in self.output_files.items():
                    print("\t", v)

        else:
            self.finished = False

        return self.output_files


def run(contig_file, output=None, quiet=False, print_files=False):
    """
    Run outside of class scope.
    """
    p = Prodigal(contig_file, output=output, quiet=quiet)
    p.run(print_files=print_files)

    return p.output_files


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="""
    A script to call Prodigal to predict both genes and proteins.
    """
    )

    parser.add_argument(
        "-i", "--input", help="Input FASTA file or dir containing fasta files"
    )
    parser.add_argument("-o", "--output", help="Path to output folder", default=".")

    args = parser.parse_args()

    if not args.input:
        parser.print_help()
        sys.exit(0)

    input_ = args.input

    @timer_wrapper
    def main():
        if os.path.isfile(input_):
            print(f"Starting script. Your input file is {input}.")
            p = Prodigal(input_, output=args.output)
            p.run()

        elif os.path.isdir(input_):

            files = os.listdir(input_)
            files = [os.path.join(input_, i) for i in files]
            files = [i for i in files if os.path.isfile(i)]

            print("\n")
            print(
                f"Starting Prodigal. You have {len(files)} files to be processed in {input_}:\n"
            )
            print("\n".join(files), "\n")

            success = 0
            failure = 0

            for contig_file in files:
                try:
                    print(f"Running Prodigal for {contig_file}.")
                    run(contig_file)
                    if os.path.isdir(os.path.splitext(contig_file)[0]):
                        success += 1
                except ValueError:
                    print(
                        f"Error for {contig_file}. Please check if it is a valid FASTA contigs file."
                    )
                    failure += 1
                    pass

            print("\n")
            print(f"Done. {success} assemblies processed. {failure} errors.")

        else:
            raise FileNotFoundError

    main()



"""
Example annotation pipeline.

1. Load contigs.

2. Run seqstats.

3. Call genes.

4. Annotate.

"""
import os
import sys
import argparse
import logging
from abacat import Genome, timer_wrapper, CONFIG

def annotate(input_, db, blast, evalue, prodigal=True):
    """
    :param input_: Input file. Must a valid FASTA contigs file (post-assembly).
    :param db: Database name. Must be in abacat.CONFIG.py db parameter.
    :param blast: Blast method. Choose from 'blastn', 'blastp' or 'blastx'. Default is 'blastn'
    :return:
    """
    logging.basicConfig(filename=os.path.splitext(input_)[0] + ".log", level = logging.INFO)
    logger = logging.getLogger(__name__)
    #handler = logging.FileHandler(os.path.splitext(input_)[0] + ".log", "w")
    #logger.addHandler(handler)
    genome = Genome(input_)
    genome.load_seqstats()
    genome.print_seqstats()
    if prodigal:
        genome.run_prodigal()
    else:
        genome.load_prodigal()
    genome.blast_seqs(db=db, blast=blast, evalue=evalue)
    handler = logging.FileHandler(os.path.join(genome.directory, genome.name + ".log"), "w")
    logger.addHandler(handler)
    #with open(os.path.join(genome.directory, genome.name + ".log"), "w") as f:
    #    f.write(handler)
    return genome


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Annotation pipeline. Starts with a contig file."
    )
    parser.add_argument(
        "-i",
        "--input",
        help="Input file. Must a valid FASTA contigs file (post-assembly).",
    )
    parser.add_argument(
        "-db",
        "--database",
        type=str,
        help="Database name. Must be in abacat.CONFIG.py db parameter.",
    )
    parser.add_argument(
        "-b",
        "--blast",
        type=str,
        default="n",
        help="Blast method. Choose from 'blastn', 'blastp' or 'blastx'. Default is 'blastn'",
    )
    parser.add_argument(
        "-e",
        "--evalue",
        type=float,
        default=CONFIG["blast"]["evalue"],
        help="E-value for BLAST. Default is the one set in abacat/config.py",
    )
    args = parser.parse_args()

    @timer_wrapper
    def run():
        annotate(args.input, args.database, args.blast, args.evalue)

    run()



CONFIG = {
    "db": "/home/ubuntu/storage/neorefs/neoref16s/rev6/neoRef16S_rev6_190325.acc",
    "out_dir": "/home/ubuntu/storage/neorefs/genomes/bacteria",
    "email": "vinicius.salazar@biome-hub.com",
    "api_key": "afb857c557897af55345e77a84def712cd08",
}



"""
Directory for data files. Contains the pathways json file and
also files for the test runs.
"""

from os import path

# This variable returns this directory
data_dir = path.dirname(__file__)
genomes_dir = path.join(data_dir, "genomes")
local_db_dir = path.join(data_dir, "db")



"""
Find phenotyping genes in a genome. Can be used with either a JSON or contigs input.
"""

import argparse
from abacat import Genome, from_json, timer_wrapper


def main(input_, evalue, json=False):
    if json:
        g = from_json(input_)
    else:
        g = Genome(input_)
        g.run_prodigal()

    g.run_pathways(evalue=evalue)
    g.to_json()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Get pathway genes from a genome. Will write json output file."
    )
    parser.add_argument(
        "-i",
        "--input",
        help="Input genome. Must be either: a valid contigs file or a json genome file.",
    )
    parser.add_argument(
        "-e",
        "--evalue",
        type=float,
        default=10 ** -3,
        help="E-value for BLAST to Pathways DB",
    )
    parser.add_argument(
        "-j",
        "--json",
        type=bool,
        default=False,
        help="Specifies that you're using an already processed JSON input.",
    )
    args = parser.parse_args()

    @timer_wrapper
    def run():
        main(args.input, args.evalue, args.json)

    run()



import setuptools

setuptools.setup(
    name="abacat",
    version="0.0.1a",
    author="Vini Salazar",
    author_email="viniws@gmail.com",
    description="abacat - A BActerial genome Curation and Annotation Toolkit",
    url="https://github.com/vinisalazar/abacat",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6"
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "biopython",
        "pandas",
        "argparse",
        "matplotlib",
        "scipy",
        "pytest"
    ]
)



"""
A script to call Prokka to do it's things. Similar to prodigal.py.

Input:
file or folder with contigs in fasta or fna format.

Output:
Prokka folder to selected output (default current directory).

Example usage:

    python prokka.py -i contigs/
    python prokka.py -i contigs.fasta

"""

import os
import sys
import time
import argparse
import datetime
import subprocess
from abacat.abacat_helper import is_fasta_wrapper, timer_wrapper


@is_fasta_wrapper
def prokka(fasta_file, output="", cpus=None, params=""):
    """
    Calls Prokka on an input file.

    Input:
    Valid FASTA file.

    Output:
    Prokka output to selected folder.

    Example:

        from prokka import prokka
        prokka("contigs.fasta", "output_folder/")
    """

    if not fasta_file:
        return "Please specify an input file."

    if output:
        output = os.path.join(os.path.abspath(output), os.path.splitext(fasta_file)[0])
    else:
        output = os.path.basename(os.path.splitext(fasta_file)[0])
    prefix = f"--prefix {output}"
    output = f"{output}_prokka"
    output_ = f"--outdir {output}"

    if not cpus:
        cpus = int(os.cpu_count() / 2)
    cpus = f"--cpus {cpus}"

    cmd = f"prokka {output_} {prefix} {fasta_file} {cpus} {params}"

    prokka = subprocess.call(cmd, shell=True)

    output_files = dict()

    for f in os.listdir(output):
        f = os.path.join(os.path.abspath(output), f)
        ext = os.path.splitext(f)[1]
        # How to change this to a more elegant dictionary?
        if ext == ".fna":
            output_files["input_contigs"] = f
        elif ext == ".faa":
            output_files["proteins"] = f
        elif ext == ".ffn":
            output_files["genes"] = f
        elif ext == ".fsa":
            output_files["submit_contigs"] = f
        elif ext == ".tbl":
            output_files["featuretable"] = f
        elif ext == ".sqn":
            output_files["sequin"] = f
        elif ext == ".gbk":
            output_files["genbank"] = f
        elif ext == ".gff":
            output_files["gff"] = f
        elif ext == ".log":
            output_files["log"] = f
        elif ext == ".txt":
            output_files["stats"] = f

    return output_files


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="""
    A script to call Prokka for genome annotation.
    """
    )

    parser.add_argument(
        "-i", "--input", help="Input FASTA file or dir containing fasta files."
    )
    parser.add_argument("-o", "--output", help="Path to output folder.", default="")
    parser.add_argument(
        "-p",
        "--params",
        help="Additional parameters to pass to Prokka as a string.",
        default="",
    )
    parser.add_argument(
        "-c",
        "--cpus",
        help="Number of CPUs to use. Default is half of the system's CPUs.",
        default=None,
    )

    args = parser.parse_args()

    if not args.input:
        parser.print_help()
        sys.exit(0)

    input = args.input

    @timer_wrapper
    def main():
        if os.path.isfile(input):
            print(f"Starting script. Your input file is {input}.")
            prokka(input, args.output, args.cpus, args.params)

        elif os.path.isdir(input):

            files = os.listdir(input)
            files = [os.path.join(input, i) for i in files]
            files = [i for i in files if os.path.isfile(i)]

            print("\n")
            print(
                f"Starting script. You have {len(files)} files to be processed in {input}:\n"
            )
            print("\n".join(files), "\n")

            success = 0
            failure = 0

            for i in files:
                try:
                    print(f"Running Prokka for {i}.")
                    prokka(i, args.output, args.cpus, args.params)
                    if os.path.isdir(os.path.splitext(i)[0]):
                        success += 1
                except Exception as err:
                    print(f"Error for {i}. Please check if it is a valid FASTA file.")
                    failure += 1
                    pass

        else:
            raise FileNotFoundError

    main()
